package ShareStuff::Actor;
use ShareStuff;
use base qw/ShareStuff/;
use strict; use warnings;
use Data::Dumper;
$Data::Dumper::Indent = 0;
use ShareStuff::Event;

use Digest::MD5 qw(md5_hex);

use OpenSRF::EX qw(:try);
use ShareStuff::AppUtils;

use Fieldmapper;
use OpenSRF::Utils::Logger qw/$logger/;
use OpenSRF::Utils qw/:datetime/;
use OpenSRF::Utils::SettingsClient;

use OpenSRF::Utils::Cache;

use OpenSRF::Utils::JSON;
use DateTime;
use DateTime::Format::ISO8601;
use ShareStuff::Const qw/:const/;

use ShareStuff::CStoreEditor qw/:funcs/;
use List::Util qw/max reduce/;

use UUID::Tiny qw/:std/;

sub initialize {
}

my $apputils = "ShareStuff::AppUtils";
my $U = $apputils;

sub _d { warn "Patron:\n" . Dumper(shift()); }

my $cache;
my $set_user_settings;
my $set_ou_settings;


#__PACKAGE__->register_method(
#	method	=> "allowed_test",
#	api_name	=> "sharestuff.actor.allowed_test",
#);
#sub allowed_test {
#    my($self, $conn, $auth, $orgid, $permcode) = @_;
#    my $e = new_editor(authtoken => $auth);
#    return $e->die_event unless $e->checkauth;
#
#    return {
#        orgid => $orgid,
#        permcode => $permcode,
#        result => $e->allowed($permcode, $orgid)
#    };
#}

__PACKAGE__->register_method(
	method	=> "update_user_setting",
	api_name	=> "sharestuff.actor.patron.settings.update",
);
sub update_user_setting {
	my($self, $conn, $auth, $user_id, $settings) = @_;
    my $e = new_editor(xact => 1, authtoken => $auth);
    return $e->die_event unless $e->checkauth;

    $user_id = $e->requestor->id unless defined $user_id;

    unless($e->requestor->id == $user_id) {
        my $user = $e->retrieve_actor_user($user_id) or return $e->die_event;
        return $e->die_event unless $e->allowed('UPDATE_USER', $user->home_ou);
    }

    for my $name (keys %$settings) {
        my $val = $$settings{$name};
        my $set = $e->search_actor_user_setting({usr => $user_id, name => $name})->[0];

        if(defined $val) {
            $val = OpenSRF::Utils::JSON->perl2JSON($val);
            if($set) {
                $set->value($val);
                $e->update_actor_user_setting($set) or return $e->die_event;
            } else {
                $set = Fieldmapper::actor::user_setting->new;
                $set->usr($user_id);
                $set->name($name);
                $set->value($val);
                $e->create_actor_user_setting($set) or return $e->die_event;
            }
        } elsif($set) {
            $e->delete_actor_user_setting($set) or return $e->die_event;
        }
    }

    $e->commit;
    return 1;
}


__PACKAGE__->register_method(
    method    => "set_ou_settings",
    api_name  => "sharestuff.actor.org_unit.settings.update",
    signature => {
        desc => "Updates the value for a given org unit setting.  The permission to update "          .
                "an org unit setting is either the UPDATE_ORG_UNIT_SETTING_ALL, or a specific "       .
                "permission specified in the update_perm column of the config.org_unit_setting_type " .
                "table's row corresponding to the setting being changed." ,
        params => [
            {desc => 'Authentication token',             type => 'string'},
            {desc => 'Org unit ID',                      type => 'number'},
            {desc => 'Hash of setting name-value pairs', type => 'object'}
        ],
        return => {desc => '1 on success, Event on error'}
    }
);

sub set_ou_settings {
	my( $self, $client, $auth, $org_id, $settings ) = @_;

    my $e = new_editor(authtoken => $auth, xact => 1);
    return $e->die_event unless $e->checkauth;

    my $all_allowed = $e->allowed("UPDATE_ORG_UNIT_SETTING_ALL", $org_id);

	for my $name (keys %$settings) {
        my $val = $$settings{$name};

        my $type = $e->retrieve_config_org_unit_setting_type([
            $name,
            {flesh => 1, flesh_fields => {'coust' => ['update_perm']}}
        ]) or return $e->die_event;
        my $set = $e->search_actor_org_unit_setting({org_unit => $org_id, name => $name})->[0];

        # If there is no relevant permission, the default assumption will
        # be, "no, the caller cannot change that value."
        return $e->die_event unless ($all_allowed ||
            ($type->update_perm && $e->allowed($type->update_perm->code, $org_id)));

        if(defined $val) {
            $val = OpenSRF::Utils::JSON->perl2JSON($val);
            if($set) {
                $set->value($val);
                $e->update_actor_org_unit_setting($set) or return $e->die_event;
            } else {
                $set = Fieldmapper::actor::org_unit_setting->new;
                $set->org_unit($org_id);
                $set->name($name);
                $set->value($val);
                $e->create_actor_org_unit_setting($set) or return $e->die_event;
            }
        } elsif($set) {
            $e->delete_actor_org_unit_setting($set) or return $e->die_event;
        }
    }

    $e->commit;
    return 1;
}

__PACKAGE__->register_method(
    method   => "user_settings",
    authoritative => 1,
    api_name => "sharestuff.actor.patron.settings.retrieve",
);
sub user_settings {
	my( $self, $client, $auth, $user_id, $setting ) = @_;

    my $e = new_editor(authtoken => $auth);
    return $e->event unless $e->checkauth;
    $user_id = $e->requestor->id unless defined $user_id;

    my $patron = $e->retrieve_actor_user($user_id) or return $e->event;
    if($e->requestor->id != $user_id) {
        return $e->event unless $e->allowed('VIEW_USER', $patron->home_ou);
    }

    sub get_setting {
        my($e, $user_id, $setting) = @_;
        my $val = $e->search_actor_user_setting({usr => $user_id, name => $setting})->[0];
        return undef unless $val; # XXX this should really return undef, but needs testing
        return OpenSRF::Utils::JSON->JSON2perl($val->value);
    }

    if($setting) {
        if(ref $setting eq 'ARRAY') {
            my %settings;
            $settings{$_} = get_setting($e, $user_id, $_) for @$setting;
            return \%settings;
        } else {
            return get_setting($e, $user_id, $setting);    
        }
    } else {
        my $s = $e->search_actor_user_setting({usr => $user_id});
	    return { map { ( $_->name => OpenSRF::Utils::JSON->JSON2perl($_->value) ) } @$s };
    }
}


__PACKAGE__->register_method(
    method    => "ranged_ou_settings",
    api_name  => "sharestuff.actor.org_unit_setting.values.ranged.retrieve",
    signature => {
        desc   => "Retrieves all org unit settings for the given org_id, up to whatever limit " .
                  "is implied for retrieving OU settings by the authenticated users' permissions.",
        params => [
            {desc => 'Authentication token',   type => 'string'},
            {desc => 'Org unit ID',            type => 'number'},
        ],
        return => {desc => 'A hashref of "ranged" settings, event on error'}
    }
);
sub ranged_ou_settings {
	my( $self, $client, $auth, $org_id ) = @_;

	my $e = new_editor(authtoken => $auth);
    return $e->event unless $e->checkauth;

    my %ranged_settings;
    my $org_list = $U->get_org_ancestors($org_id);
    my $settings = $e->search_actor_org_unit_setting({org_unit => $org_list});
    $org_list = [ reverse @$org_list ];

    # start at the context org and capture the setting value
    # without clobbering settings we've already captured
    for my $this_org_id (@$org_list) {
        
        my @sets = grep { $_->org_unit == $this_org_id } @$settings;

        for my $set (@sets) {
            my $type = $e->retrieve_config_org_unit_setting_type([
                $set->name,
                {flesh => 1, flesh_fields => {coust => ['view_perm']}}
            ]);

            # If there is no relevant permission, the default assumption will
            # be, "yes, the caller can have that value."
            if ($type && $type->view_perm) {
                next if not $e->allowed($type->view_perm->code, $org_id);
            }

            $ranged_settings{$set->name} = OpenSRF::Utils::JSON->JSON2perl($set->value)
                unless defined $ranged_settings{$set->name};
        }
    }

	return \%ranged_settings;
}



__PACKAGE__->register_method(
    api_name  => 'sharestuff.actor.ou_setting.ancestor_default',
    method    => 'ou_ancestor_setting',
    signature => {
        desc => 'Get the org unit setting value associated with the setting name as seen from the specified org unit.  ' .
                'IF AND ONLY IF an authentication token is provided, this method will make sure that the given '         .
                'user has permission to view that setting, if there is a permission associated with the setting.'        ,
        params => [
            { desc => 'Org unit ID',          type => 'number' },
            { desc => 'setting name',         type => 'string' },
            { desc => 'authtoken (optional)', type => 'string' }
        ],
        return => {desc => 'A value for the org unit setting, or undef'}
    }
);

# ------------------------------------------------------------------
# Attempts to find the org setting value for a given org.  if not 
# found at the requested org, searches up the org tree until it 
# finds a parent that has the requested setting.
# when found, returns { org => $id, value => $value }
# otherwise, returns NULL
# ------------------------------------------------------------------
sub ou_ancestor_setting {
    my( $self, $client, $orgid, $name, $auth ) = @_;
    return $U->ou_ancestor_setting($orgid, $name, undef, $auth);
}

__PACKAGE__->register_method(
    api_name  => 'sharestuff.actor.ou_setting.ancestor_default.batch',
    method    => 'ou_ancestor_setting_batch',
    signature => {
        desc => 'Get org unit setting name => value pairs for a list of names, as seen from the specified org unit.  ' .
                'IF AND ONLY IF an authentication token is provided, this method will make sure that the given '       .
                'user has permission to view that setting, if there is a permission associated with the setting.'      ,
        params => [
            { desc => 'Org unit ID',          type => 'number' },
            { desc => 'setting name list',    type => 'array'  },
            { desc => 'authtoken (optional)', type => 'string' }
        ],
        return => {desc => 'A hash with name => value pairs for the org unit settings'}
    }
);
sub ou_ancestor_setting_batch {
    my( $self, $client, $orgid, $name_list, $auth ) = @_;
    my %values;
    $values{$_} = $U->ou_ancestor_setting($orgid, $_, undef, $auth) for @$name_list;
    return \%values;
}



__PACKAGE__->register_method(
    method   => "update_patron",
    api_name => "sharestuff.actor.patron.update",
    signature => {
        desc   => q/
            Update an existing user, or create a new one.  Related objects,
            like cards, addresses, survey responses, and stat cats, 
            can be updated by attaching them to the user object in their
            respective fields.  For examples, the billing address object
            may be inserted into the 'billing_address' field, etc.  For each 
            attached object, indicate if the object should be created, 
            updated, or deleted using the built-in 'isnew', 'ischanged', 
            and 'isdeleted' fields on the object.
        /,
        params => [
            { desc => 'Authentication token', type => 'string' },
            { desc => 'Patron data object',   type => 'object' }
        ],
        return => {desc => 'A fleshed user object, event on error'}
    }
);

sub update_patron {
	my( $self, $client, $user_session, $patron ) = @_;

	my $session = $apputils->start_db_session();

	$logger->info($patron->isnew ? "Creating new patron..." : "Updating Patron: " . $patron->id);

	my( $user_obj, $evt ) = $U->checkses($user_session);
	return $evt if $evt;

	$evt = check_group_perm($session, $user_obj, $patron);
	return $evt if $evt;


	# $new_patron is the patron in progress.  $patron is the original patron
	# passed in with the method.  new_patron will change as the components
	# of patron are added/updated.

	my $new_patron;

	# create/update the patron first so we can use his id

    # $patron is the obj from the client (new data) and $new_patron is the
    # patron object properly built for db insertion, so we need a third variable
    # if we want to represent the old patron.

    my $old_patron;
    my $barred_hook = '';

	if($patron->isnew()) {
		( $new_patron, $evt ) = _add_patron($session, _clone_patron($patron), $user_obj);
		return $evt if $evt;
		if($U->is_true($patron->barred)) {
			$evt = $U->check_perms($user_obj->id, $patron->home_ou, 'BAR_PATRON');
			return $evt if $evt;
		}
	} else {
        $new_patron = $patron;

        # Did auth checking above already.
        my $e = new_editor;
        $old_patron = $e->retrieve_actor_user($patron->id) or
            return $e->die_event;
        $e->disconnect;
        if($U->is_true($old_patron->barred) != $U->is_true($new_patron->barred)) {
            $evt = $U->check_perms($user_obj->id, $patron->home_ou, $U->is_true($old_patron->barred) ? 'UNBAR_PATRON' : 'BAR_PATRON');
            return $evt if $evt;

            $barred_hook = $U->is_true($new_patron->barred) ? 
                'au.barred' : 'au.unbarred';
        }
    }

	($new_patron, $evt) = _create_perm_maps($session, $user_session, $patron, $new_patron, $user_obj);
	return $evt if $evt;

	$apputils->commit_db_session($session);

	return flesh_user($new_patron->id(), new_editor(requestor => $user_obj, xact => 1));
}


sub flesh_user {
	my $id = shift;
    my $e = shift;
    my $home_ou = shift;

    my $fields = [ "settings" ];
    push @$fields, "home_ou" if $home_ou;
    my $u = $e->retrieve_actor_user(
    [
        $id,
        {
            "flesh"             => 1,
            "flesh_fields" =>  { "au" => $fields }
        }
    ]
    ) or return $e->die_event;

    return $u;
}






# clone and clear stuff that would break the database
sub _clone_patron {
	my $patron = shift;

	my $new_patron = $patron->clone;
	# clear these
	$new_patron->clear_id();
	$new_patron->clear_isnew();
	$new_patron->clear_ischanged();
	$new_patron->clear_isdeleted();
	$new_patron->clear_permissions();

	return $new_patron;
}


# XXX move to cstore...
sub _add_patron {

	my $session		= shift;
	my $patron		= shift;
	my $user_obj	= shift;

	my $evt = $U->check_perms($user_obj->id, $patron->home_ou, 'CREATE_USER');
	return (undef, $evt) if $evt;

	my $ex = $session->request(
		'sharestuff.cstore.direct.actor.user.search', { 'usrname' => $patron->usrname() })->gather(1);
	if( $ex and @$ex ) {
		return (undef, ShareStuff::Event->new('USERNAME_EXISTS'));
	}

	$logger->info("Creating new user in the DB with username: ".$patron->usrname());

	my $usr = $session->request(
		"sharestuff.cstore.direct.actor.user.create", $patron)->gather(1);
	return (undef, $U->DB_UPDATE_FAILED($patron)) unless $id;

	$logger->info("Successfully created new user [".$usr->id."] in DB");

	return ( $usr, undef );
}


sub check_group_perm {
	my( $session, $requestor, $patron ) = @_;
	my $evt;

	# first let's see if the requestor has 
	# priveleges to update this user in any way
	if( ! $patron->isnew ) {
		my $p = $session->request(
			'sharestuff.cstore.direct.actor.user.retrieve', $patron->id )->gather(1);

		# If we are the requestor (trying to update our own account)
		# and we are not trying to change our profile, we're good
		if( $p->id == $requestor->id and 
				$p->profile == $patron->profile ) {
			return undef;
		}


		$evt = group_perm_failed($session, $requestor, $p);
		return $evt if $evt;
	}

	# They are allowed to edit this patron.. can they put the 
	# patron into the group requested?
	$evt = group_perm_failed($session, $requestor, $patron);
	return $evt if $evt;
	return undef;
}


sub group_perm_failed {
	my( $session, $requestor, $patron ) = @_;

	my $perm;
	my $grp;
	my $grpid = $patron->profile;

	do {

		$logger->debug("user update looking for group perm for group $grpid");
		$grp = $session->request(
			'sharestuff.cstore.direct.permission.grp_tree.retrieve', $grpid )->gather(1);
		return ShareStuff::Event->new('PERMISSION_GRP_TREE_NOT_FOUND') unless $grp;

	} while( !($perm = $grp->application_perm) and ($grpid = $grp->parent) );

	$logger->info("user update checking perm $perm on user ".
		$requestor->id." for update/create on user username=".$patron->usrname);

	my $evt = $U->check_perms($requestor->id, $patron->home_ou, $perm);
	return $evt if $evt;
	return undef;
}



sub _update_patron {
	my( $session, $patron, $user_obj, $noperm) = @_;

	$logger->info("Updating patron ".$patron->id." in DB");

	my $evt;

	if(!$noperm) {
		$evt = $U->check_perms($user_obj->id, $patron->home_ou, 'UPDATE_USER');
		return (undef, $evt) if $evt;
	}

	if(!$patron->ident_type) {
		$patron->clear_ident_type;
		$patron->clear_ident_value;
	}

	my $stat = $session->request(
		"sharestuff.cstore.direct.actor.user.update",$patron )->gather(1);
	return (undef, $U->DB_UPDATE_FAILED($patron)) unless defined($stat);

	return ($patron);
}

sub verify_last_xact {
    my( $session, $patron ) = @_;
    return undef unless $patron->id and $patron->id > 0;
    my $p = $session->request(
        'sharestuff.cstore.direct.actor.user.retrieve', $patron->id)->gather(1);
    my $xact = $p->last_xact_id;
    return undef unless $xact;
    $logger->info("user xact = $xact, saving with xact " . $patron->last_xact_id);
    return ShareStuff::Event->new('XACT_COLLISION')
        if $xact ne $patron->last_xact_id;
    return undef;
}



sub _create_stat_maps {

	my($session, $user_session, $patron, $new_patron) = @_;

	my $maps = $patron->stat_cat_entries();

	for my $map (@$maps) {

		my $method = "sharestuff.cstore.direct.actor.stat_cat_entry_user_map.update";

		if ($map->isdeleted()) {
			$method = "sharestuff.cstore.direct.actor.stat_cat_entry_user_map.delete";

		} elsif ($map->isnew()) {
			$method = "sharestuff.cstore.direct.actor.stat_cat_entry_user_map.create";
			$map->clear_id;
		}


		$map->target_usr($new_patron->id);

		#warn "
		$logger->info("Updating stat entry with method $method and map $map");

		my $stat = $session->request($method, $map)->gather(1);
		return (undef, $U->DB_UPDATE_FAILED($map)) unless defined($stat);

	}

	return ($new_patron, undef);
}

sub _create_perm_maps {

	my($session, $user_session, $patron, $new_patron) = @_;

	my $maps = $patron->permissions;

	for my $map (@$maps) {

		my $method = "sharestuff.cstore.direct.permission.usr_perm_map.update";
		if ($map->isdeleted()) {
			$method = "sharestuff.cstore.direct.permission.usr_perm_map.delete";
		} elsif ($map->isnew()) {
			$method = "sharestuff.cstore.direct.permission.usr_perm_map.create";
			$map->clear_id;
		}


		$map->usr($new_patron->id);

		#warn( "Updating permissions with method $method and session $user_session and map $map" );
		$logger->info( "Updating permissions with method $method and map $map" );

		my $stat = $session->request($method, $map)->gather(1);
		return (undef, $U->DB_UPDATE_FAILED($map)) unless defined($stat);

	}

	return ($new_patron, undef);
}


__PACKAGE__->register_method(
    method   => "set_user_perms",
    api_name => "sharestuff.actor.user.permissions.update",
);

sub set_user_perms {
	my $self = shift;
	my $client = shift;
	my $ses = shift;
	my $maps = shift;

	my $session = $apputils->start_db_session();

	my( $user_obj, $evt ) = $U->checkses($ses);
	return $evt if $evt;

	my $all = undef;
	$all = 1 if ($U->is_true($user_obj->super_user()));
    $all = 1 unless ($U->check_perms($user_obj->id, $user_obj->home_ou, 'EVERYTHING'));

	for my $map (@$maps) {

		my $method = "sharestuff.cstore.direct.permission.usr_perm_map.update";
		if ($map->isdeleted()) {
			$method = "sharestuff.cstore.direct.permission.usr_perm_map.delete";
		} elsif ($map->isnew()) {
			$method = "sharestuff.cstore.direct.permission.usr_perm_map.create";
			$map->clear_id;
		}

		next if (!$all);
		#warn( "Updating permissions with method $method and session $ses and map $map" );
		$logger->info( "Updating permissions with method $method and map $map" );

		my $stat = $session->request($method, $map)->gather(1);
		$logger->warn( "update failed: ".$U->DB_UPDATE_FAILED($map) ) unless defined($stat);

	}

	$apputils->commit_db_session($session);

	return scalar(@$maps);
}


__PACKAGE__->register_method(
	method	=> "user_retrieve_by_barcode",
    authoritative => 1,
	api_name	=> "sharestuff.actor.user.fleshed.retrieve_by_barcode",);

sub user_retrieve_by_barcode {
	my($self, $client, $auth, $barcode, $flesh_home_ou) = @_;

    my $e = new_editor(authtoken => $auth);
    return $e->event unless $e->checkauth;

    my $card = $e->search_actor_card({barcode => $barcode})->[0]
        or return $e->event;

	my $user = flesh_user($card->usr, $e, $flesh_home_ou);
    return $e->event unless $e->allowed(
        "VIEW_USER", $flesh_home_ou ? $user->home_ou->id : $user->home_ou
    );
    return $user;
}



__PACKAGE__->register_method(
    method        => "get_user_by_id",
    authoritative => 1,
    api_name      => "sharestuff.actor.user.retrieve",
);

sub get_user_by_id {
	my ($self, $client, $auth, $id) = @_;
	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;
	my $user = $e->retrieve_actor_user($id) or return $e->event;
	return $e->event unless $e->allowed('VIEW_USER', $user->home_ou);	
	return $user;
}


__PACKAGE__->register_method(
    method   => "get_org_types",
    api_name => "sharestuff.actor.org_types.retrieve",
);
sub get_org_types {
    return $U->get_org_types();
}


__PACKAGE__->register_method(
    method   => "get_user_ident_types",
    api_name => "sharestuff.actor.user.ident_types.retrieve",
);
my $ident_types;
sub get_user_ident_types {
	return $ident_types if $ident_types;
	return $ident_types = 
		new_editor()->retrieve_all_config_identification_type();
}


__PACKAGE__->register_method(
    method   => "get_org_unit",
    api_name => "sharestuff.actor.org_unit.retrieve",
);

sub get_org_unit {
	my( $self, $client, $user_session, $org_id ) = @_;
	my $e = new_editor(authtoken => $user_session);
	if(!$org_id) {
		return $e->event unless $e->checkauth;
		$org_id = $e->requestor->ws_ou;
	}
	my $o = $e->retrieve_actor_org_unit($org_id)
		or return $e->event;
	return $o;
}

__PACKAGE__->register_method(
    method   => "search_org_unit",
    api_name => "sharestuff.actor.org_unit_list.search",
);

sub search_org_unit {

	my( $self, $client, $field, $value ) = @_;

	my $list = ShareStuff::AppUtils->simple_scalar_request(
		"sharestuff.cstore",
		"sharestuff.cstore.direct.actor.org_unit.search.atomic", 
		{ $field => $value } );

	return $list;
}


# build the org tree

__PACKAGE__->register_method(
	method	=> "get_org_tree",
	api_name	=> "sharestuff.actor.org_tree.retrieve",
	argc		=> 0, 
	note		=> "Returns the entire org tree structure",
);

sub get_org_tree {
	my $self = shift;
	my $client = shift;
	return $U->get_org_tree($client->session->session_locale);
}


__PACKAGE__->register_method(
    method    => "update_passwd",
    api_name  => "sharestuff.actor.user.password.update",
    signature => {
        desc   => "Update the operator's password", 
        params => [
            { desc => 'Authentication token', type => 'string' },
            { desc => 'New password',         type => 'string' },
            { desc => 'Current password',     type => 'string' }
        ],
        return => {desc => '1 on success, Event on error or incorrect current password'}
    }
);

__PACKAGE__->register_method(
    method    => "update_passwd",
    api_name  => "sharestuff.actor.user.username.update",
    signature => {
        desc   => "Update the operator's username", 
        params => [
            { desc => 'Authentication token', type => 'string' },
            { desc => 'New username',         type => 'string' },
            { desc => 'Current password',     type => 'string' }
        ],
        return => {desc => '1 on success, Event on error or incorrect current password'}
    }
);

__PACKAGE__->register_method(
    method    => "update_passwd",
    api_name  => "sharestuff.actor.user.email.update",
    signature => {
        desc   => "Update the operator's email address", 
        params => [
            { desc => 'Authentication token', type => 'string' },
            { desc => 'New email address',    type => 'string' },
            { desc => 'Current password',     type => 'string' }
        ],
        return => {desc => '1 on success, Event on error or incorrect current password'}
    }
);

sub update_passwd {
    my( $self, $conn, $auth, $new_val, $orig_pw ) = @_;
    my $e = new_editor(xact=>1, authtoken=>$auth);
    return $e->die_event unless $e->checkauth;

    my $db_user = $e->retrieve_actor_user($e->requestor->id)
        or return $e->die_event;
    my $api = $self->api_name;

    # make sure the original password matches the in-database password
    if (md5_hex($orig_pw) ne $db_user->passwd) {
        $e->rollback;
        return new ShareStuff::Event('INCORRECT_PASSWORD');
    }

    if( $api =~ /password/o ) {

        $db_user->passwd($new_val);

    } else {

        # if we don't clear the password, the user will be updated with
        # a hashed version of the hashed version of their password
        $db_user->clear_passwd;

        if( $api =~ /username/o ) {

            # make sure no one else has this username
            my $exist = $e->search_actor_user({usrname=>$new_val},{idlist=>1}); 
            if (@$exist) {
                $e->rollback;
                return new ShareStuff::Event('USERNAME_EXISTS');
            }
            $db_user->usrname($new_val);

        } elsif( $api =~ /email/o ) {
            $db_user->email($new_val);
        }
    }

    $e->update_actor_user($db_user) or return $e->die_event;
    $e->commit;

    # update the cached user to pick up these changes
    $U->simplereq('sharestuff.auth', 'sharestuff.auth.session.reset_timeout', $auth, 1);
    return 1;
}



__PACKAGE__->register_method(
    method   => "check_user_perms",
    api_name => "sharestuff.actor.user.perm.check",
    notes    => <<"	NOTES");
	Takes a login session, user id, an org id, and an array of perm type strings.  For each
	perm type, if the user does *not* have the given permission it is added
	to a list which is returned from the method.  If all permissions
	are allowed, an empty list is returned
	if the logged in user does not match 'user_id', then the logged in user must
	have VIEW_PERMISSION priveleges.
	NOTES

sub check_user_perms {
	my( $self, $client, $login_session, $user_id, $org_id, $perm_types ) = @_;

	my( $staff, $evt ) = $apputils->checkses($login_session);
	return $evt if $evt;

	if($staff->id ne $user_id) {
		if( $evt = $apputils->check_perms(
			$staff->id, $org_id, 'VIEW_PERMISSION') ) {
			return $evt;
		}
	}

	my @not_allowed;
	for my $perm (@$perm_types) {
		if($apputils->check_perms($user_id, $org_id, $perm)) {
			push @not_allowed, $perm;
		}
	}

	return \@not_allowed
}

__PACKAGE__->register_method(
	method	=> "check_user_perms2",
	api_name	=> "sharestuff.actor.user.perm.check.multi_org",
	notes		=> q/
		Checks the permissions on a list of perms and orgs for a user
		@param authtoken The login session key
		@param user_id The id of the user to check
		@param orgs The array of org ids
		@param perms The array of permission names
		@return An array of  [ orgId, permissionName ] arrays that FAILED the check
		if the logged in user does not match 'user_id', then the logged in user must
		have VIEW_PERMISSION priveleges.
	/);

sub check_user_perms2 {
	my( $self, $client, $authtoken, $user_id, $orgs, $perms ) = @_;

	my( $staff, $target, $evt ) = $apputils->checkses_requestor(
		$authtoken, $user_id, 'VIEW_PERMISSION' );
	return $evt if $evt;

	my @not_allowed;
	for my $org (@$orgs) {
		for my $perm (@$perms) {
			if($apputils->check_perms($user_id, $org, $perm)) {
				push @not_allowed, [ $org, $perm ];
			}
		}
	}

	return \@not_allowed
}


__PACKAGE__->register_method(
	method => 'check_user_perms3',
	api_name	=> 'sharestuff.actor.user.perm.highest_org',
	notes		=> q/
		Returns the highest org unit id at which a user has a given permission
		If the requestor does not match the target user, the requestor must have
		'VIEW_PERMISSION' rights at the home org unit of the target user
		@param authtoken The login session key
		@param userid The id of the user in question
		@param perm The permission to check
		@return The org unit highest in the org tree within which the user has
		the requested permission
	/);

sub check_user_perms3 {
	my($self, $client, $authtoken, $user_id, $perm) = @_;
	my $e = new_editor(authtoken=>$authtoken);
	return $e->event unless $e->checkauth;

	my $tree = $U->get_org_tree();

    unless($e->requestor->id == $user_id) {
        my $user = $e->retrieve_actor_user($user_id)
            or return $e->event;
        return $e->event unless $e->allowed('VIEW_PERMISSION', $user->home_ou);
	    return $U->find_highest_perm_org($perm, $user_id, $user->home_ou, $tree );
    }

    return $U->find_highest_perm_org($perm, $user_id, $e->requestor->ws_ou, $tree);
}

__PACKAGE__->register_method(
	method => 'user_has_work_perm_at',
	api_name	=> 'sharestuff.actor.user.has_work_perm_at',
    authoritative => 1,
    signature => {
        desc => q/
            Returns a set of org unit IDs which represent the highest orgs in 
            the org tree where the user has the requested permission.  The
            purpose of this method is to return the smallest set of org units
            which represent the full expanse of the user's ability to perform
            the requested action.  The user whose perms this method should
            check is implied by the authtoken. /,
        params => [
		    {desc => 'authtoken', type => 'string'},
            {desc => 'permission name', type => 'string'},
            {desc => q/user id, optional.  If present, check perms for 
                this user instead of the logged in user/, type => 'number'},
        ],
        return => {desc => 'An array of org IDs'}
    }
);

sub user_has_work_perm_at {
    my($self, $conn, $auth, $perm, $user_id) = @_;
    my $e = new_editor(authtoken=>$auth);
    return $e->event unless $e->checkauth;
    if(defined $user_id) {
        my $user = $e->retrieve_actor_user($user_id) or return $e->event;
        return $e->event unless $e->allowed('VIEW_PERMISSION', $user->home_ou);
    }
    return $U->user_has_work_perm_at($e, $perm, undef, $user_id);
}

__PACKAGE__->register_method(
	method => 'user_has_work_perm_at_batch',
	api_name	=> 'sharestuff.actor.user.has_work_perm_at.batch',
    authoritative => 1,
);

sub user_has_work_perm_at_batch {
    my($self, $conn, $auth, $perms, $user_id) = @_;
    my $e = new_editor(authtoken=>$auth);
    return $e->event unless $e->checkauth;
    if(defined $user_id) {
        my $user = $e->retrieve_actor_user($user_id) or return $e->event;
        return $e->event unless $e->allowed('VIEW_PERMISSION', $user->home_ou);
    }
    my $map = {};
    $map->{$_} = $U->user_has_work_perm_at($e, $_) for @$perms;
    return $map;
}



__PACKAGE__->register_method(
	method => 'check_user_perms4',
	api_name	=> 'sharestuff.actor.user.perm.highest_org.batch',
	notes		=> q/
		Returns the highest org unit id at which a user has a given permission
		If the requestor does not match the target user, the requestor must have
		'VIEW_PERMISSION' rights at the home org unit of the target user
		@param authtoken The login session key
		@param userid The id of the user in question
		@param perms An array of perm names to check 
		@return An array of orgId's  representing the org unit 
		highest in the org tree within which the user has the requested permission
		The arrah of orgId's has matches the order of the perms array
	/);

sub check_user_perms4 {
	my( $self, $client, $authtoken, $userid, $perms ) = @_;
	
	my( $staff, $target, $org, $evt );

	( $staff, $target, $evt ) = $apputils->checkses_requestor(
		$authtoken, $userid, 'VIEW_PERMISSION' );
	return $evt if $evt;

	my @arr;
	return [] unless ref($perms);
	my $tree = $U->get_org_tree();

	for my $p (@$perms) {
		push( @arr, $U->find_highest_perm_org( $p, $userid, $target->home_ou, $tree ) );
	}
	return \@arr;
}


__PACKAGE__->register_method(
    method        => "user_fines_summary",
    api_name      => "sharestuff.actor.user.fines.summary",
    authoritative => 1,
    signature     => {
        desc   => 'Returns a short summary of the users total open fines, '  .
                  'excluding voided fines Params are login_session, user_id' ,
        params => [
            {desc => 'Authentication token', type => 'string'},
            {desc => 'User ID',              type => 'string'}  # number?
        ],
        return => {
            desc => "a 'mous' object, event on error",
        }
    }
);

sub user_fines_summary {
	my( $self, $client, $auth, $user_id ) = @_;

	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;

	if( $user_id ne $e->requestor->id ) {
	    my $user = $e->retrieve_actor_user($user_id) or return $e->event;
		return $e->event unless 
			$e->allowed('VIEW_USER_FINES_SUMMARY', $user->home_ou);
	}

    return $e->search_money_open_user_summary({usr => $user_id})->[0];
}


__PACKAGE__->register_method(
    method        => "user_opac_vitals",
    api_name      => "sharestuff.actor.user.opac.vital_stats",
    argc          => 1,
    authoritative => 1,
    signature     => {
        desc   => 'Returns a short summary of the users vital stats, including '  .
                  'identification information, accumulated balance, number of holds, ' .
                  'and current open circulation stats' ,
        params => [
            {desc => 'Authentication token',                          type => 'string'},
            {desc => 'Optional User ID, for use in the staff client', type => 'number'}  # number?
        ],
        return => {
            desc => "An object with four properties: user, fines, checkouts and holds."
        }
    }
);

sub user_opac_vitals {
	my( $self, $client, $auth, $user_id ) = @_;

	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;

    $user_id ||= $e->requestor->id;

    my $user = $e->retrieve_actor_user( $user_id );

    my ($fines) = $self
        ->method_lookup('sharestuff.actor.user.fines.summary')
        ->run($auth => $user_id);
    return $fines if (defined($U->event_code($fines)));

    if (!$fines) {
        $fines = new Fieldmapper::money::open_user_summary ();
        $fines->balance_owed(0.00);
        $fines->total_owed(0.00);
        $fines->total_paid(0.00);
        $fines->usr($user_id);
    }

    my ($holds) = $self
        ->method_lookup('sharestuff.actor.user.hold_requests.count')
        ->run($auth => $user_id);
    return $holds if (defined($U->event_code($holds)));

    my ($out) = $self
        ->method_lookup('sharestuff.actor.user.checked_out.count')
        ->run($auth => $user_id);
    return $out if (defined($U->event_code($out)));

    $out->{"total_out"} = reduce { $a + $out->{$b} } 0, qw/out overdue long_overdue/;

    return {
        user => {
            first_given_name  => $user->first_given_name,
            second_given_name => $user->second_given_name,
            family_name       => $user->family_name,
            alias             => $user->alias,
            usrname           => $user->usrname
        },
        fines => $fines->to_bare_hash,
        checkouts => $out,
        holds => $holds
    };
}


##### a small consolidation of related method registrations
my $common_params = [
    { desc => 'Authentication token', type => 'string' },
    { desc => 'User ID',              type => 'string' },
    { desc => 'Transactions type (optional, defaults to all)', type => 'string' },
    { desc => 'Options hash.  May contain limit and offset for paged results.', type => 'object' },
];
my %methods = (
    'sharestuff.actor.user.transactions'                      => '',
    'sharestuff.actor.user.transactions.fleshed'              => '',
    'sharestuff.actor.user.transactions.have_charge'          => ' that have an initial charge',
    'sharestuff.actor.user.transactions.have_charge.fleshed'  => ' that have an initial charge',
    'sharestuff.actor.user.transactions.have_balance'         => ' that have an outstanding balance',
    'sharestuff.actor.user.transactions.have_balance.fleshed' => ' that have an outstanding balance',
);

foreach (keys %methods) {
    my %args = (
        method    => "user_transactions",
        api_name  => $_,
        signature => {
            desc   => 'For a given user, retrieve a list of '
                    . (/\.fleshed/ ? 'fleshed ' : '')
                    . 'transactions' . $methods{$_}
                    . ' optionally limited to transactions of a given type.',
            params => $common_params,
            return => {
                desc => "List of objects, or event on error.  Each object is a hash containing: transaction, circ, record. "
                      . 'These represent the relevant (mbts) transaction, attached circulation and title pointed to in the circ, respectively.',
            }
        }
    );
    $args{authoritative} = 1;
    __PACKAGE__->register_method(%args);
}

# Now for the counts
%methods = (
    'sharestuff.actor.user.transactions.count'              => '',
    'sharestuff.actor.user.transactions.have_charge.count'  => ' that have an initial charge',
    'sharestuff.actor.user.transactions.have_balance.count' => ' that have an outstanding balance',
);

foreach (keys %methods) {
    my %args = (
        method    => "user_transactions",
        api_name  => $_,
        signature => {
            desc   => 'For a given user, retrieve a count of open '
                    . 'transactions' . $methods{$_}
                    . ' optionally limited to transactions of a given type.',
            params => $common_params,
            return => { desc => "Integer count of transactions, or event on error" }
        }
    );
    /\.have_balance/ and $args{authoritative} = 1;     # FIXME: I don't know why have_charge isn't authoritative
    __PACKAGE__->register_method(%args);
}

__PACKAGE__->register_method(
    method        => "user_transactions",
    api_name      => "sharestuff.actor.user.transactions.have_balance.total",
    authoritative => 1,
    signature     => {
        desc   => 'For a given user, retrieve the total balance owed for open transactions,'
                . ' optionally limited to transactions of a given type.',
        params => $common_params,
        return => { desc => "Decimal balance value, or event on error" }
    }
);


sub user_transactions {
	my( $self, $client, $auth, $user_id, $type, $options ) = @_;
    $options ||= {};

    my $e = new_editor(authtoken => $auth);
    return $e->event unless $e->checkauth;

    my $user = $e->retrieve_actor_user($user_id) or return $e->event;

    return $e->event unless 
        $e->requestor->id == $user_id or
        $e->allowed('VIEW_USER_TRANSACTIONS', $user->home_ou);

    my $api = $self->api_name();

    my $filter = ($api =~ /have_balance/o) ?
        { 'balance_owed' => { '<>' => 0 } }:
        { 'total_owed' => { '>' => 0 } };

    my $method = 'sharestuff.actor.user.transactions.history.still_open';
    $method = "$method.authoritative" if $api =~ /authoritative/;
    my ($trans) = $self->method_lookup($method)->run($auth, $user_id, $type, $filter, $options);

	if($api =~ /total/o) { 
		my $total = 0.0;
        $total += $_->balance_owed for @$trans;
		return $total;
	}

    ($api =~ /count/o  ) and return scalar @$trans;
    ($api !~ /fleshed/o) and return $trans;

	my @resp;
	for my $t (@$trans) {
			
		if( $t->xact_type ne 'circulation' ) {
			push @resp, {transaction => $t};
			next;
		}

        my $circ_data = flesh_circ($e, $t->id);
		push @resp, {transaction => $t, %$circ_data};
	}

	return \@resp; 
} 


__PACKAGE__->register_method(
    method   => "user_transaction_retrieve",
    api_name => "sharestuff.actor.user.transaction.fleshed.retrieve",
    argc     => 1,
    authoritative => 1,
    notes    => "Returns a fleshed transaction record"
);

__PACKAGE__->register_method(
    method   => "user_transaction_retrieve",
    api_name => "sharestuff.actor.user.transaction.retrieve",
    argc     => 1,
    authoritative => 1,
    notes    => "Returns a transaction record"
);

sub user_transaction_retrieve {
	my($self, $client, $auth, $bill_id) = @_;

    my $e = new_editor(authtoken => $auth);
    return $e->event unless $e->checkauth;

    my $trans = $e->retrieve_money_billable_transaction_summary(
        [$bill_id, {flesh => 1, flesh_fields => {mbts => ['usr']}}]) or return $e->event;

    return $e->event unless $e->allowed('VIEW_USER_TRANSACTIONS', $trans->usr->home_ou);

    $trans->usr($trans->usr->id); # de-flesh for backwards compat

    return $trans unless $self->api_name =~ /flesh/;
    return {transaction => $trans} if $trans->xact_type ne 'circulation';

    my $circ_data = flesh_circ($e, $trans->id, 1);

	return {transaction => $trans, %$circ_data};
}

sub flesh_circ {
    my $e = shift;
    my $circ_id = shift;
    my $flesh_copy = shift;

    my $circ = $e->retrieve_action_circulation([
        $circ_id, {
            flesh => 3,
            flesh_fields => {
                circ => ['target_copy'],
                acp => ['call_number'],
                acn => ['record']
            }
        }
    ]);

	my $mods;
    my $copy = $circ->target_copy;

    if($circ->target_copy->call_number->id == OILS_PRECAT_CALL_NUMBER) {
        $mods = new Fieldmapper::metabib::virtual_record;
        $mods->doc_id(OILS_PRECAT_RECORD);
        $mods->title($copy->dummy_title);
        $mods->author($copy->dummy_author);

    } else {
        $mods = $U->record_to_mvr($circ->target_copy->call_number->record);
    }

    # more de-fleshiing
    $circ->target_copy($circ->target_copy->id);
    $copy->call_number($copy->call_number->id);

	return {circ => $circ, record => $mods, copy => ($flesh_copy) ? $copy : undef };
}


__PACKAGE__->register_method(
    method        => "hold_request_count",
    api_name      => "sharestuff.actor.user.hold_requests.count",
    authoritative => 1,
    argc          => 1,
    notes         => 'Returns hold ready/total counts'
);
	
sub hold_request_count {
	my( $self, $client, $authtoken, $user_id ) = @_;
    my $e = new_editor(authtoken => $authtoken);
    return $e->event unless $e->checkauth;

    $user_id = $e->requestor->id unless defined $user_id;

    if($e->requestor->id ne $user_id) {
        my $user = $e->retrieve_actor_user($user_id);
        return $e->event unless $e->allowed('VIEW_HOLD', $user->home_ou);
    }

    my $holds = $e->json_query({
        select => {ahr => ['pickup_lib', 'current_shelf_lib']},
        from => 'ahr',
        where => {
            usr => $user_id,
            fulfillment_time => {"=" => undef },
            cancel_time => undef,
        }
    });

	return { 
        total => scalar(@$holds), 
        ready => scalar(
            grep { 
                $_->{current_shelf_lib} and # avoid undef warnings
                $_->{pickup_lib} eq $_->{current_shelf_lib} 
            } @$holds
        ) 
    };
}

__PACKAGE__->register_method(
    method        => "checked_out",
    api_name      => "sharestuff.actor.user.checked_out",
    authoritative => 1,
    argc          => 2,
	signature     => {
        desc => "For a given user, returns a structure of circulations objects sorted by out, overdue, lost, claims_returned, long_overdue. "
              . "A list of IDs are returned of each type.  Circs marked lost, long_overdue, and claims_returned will not be 'finished' "
              . "(i.e., outstanding balance or some other pending action on the circ). "
              . "The .count method also includes a 'total' field which sums all open circs.",
        params => [
            { desc => 'Authentication Token', type => 'string'},
            { desc => 'User ID',              type => 'string'},
        ],
        return => {
            desc => 'Returns event on error, or an object with ID lists, like: '
                  . '{"out":[12552,451232], "claims_returned":[], "long_overdue":[23421] "overdue":[], "lost":[]}'
        },
    }
);

__PACKAGE__->register_method(
    method        => "checked_out",
    api_name      => "sharestuff.actor.user.checked_out.count",
    authoritative => 1,
    argc          => 2,
    signature     => q/@see sharestuff.actor.user.checked_out/
);

sub checked_out {
	my( $self, $conn, $auth, $userid ) = @_;

	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;

	if( $userid ne $e->requestor->id ) {
        my $user = $e->retrieve_actor_user($userid) or return $e->event;
		return $e->event unless($e->allowed('VIEW_CIRCULATIONS', $user->home_ou));
	}

	my $count = $self->api_name =~ /count/;
	return _checked_out( $count, $e, $userid );
}

sub _checked_out {
	my( $iscount, $e, $userid ) = @_;

    my %result = (
        out => [],
        overdue => [],
        lost => [],
        claims_returned => [],
        long_overdue => []
    );
	my $meth = 'retrieve_action_open_circ_';

    if ($iscount) {
	    $meth .= 'count';
        %result = (
            out => 0,
            overdue => 0,
            lost => 0,
            claims_returned => 0,
            long_overdue => 0
        );
    } else {
	    $meth .= 'list';
    }

    my $data = $e->$meth($userid);

    if ($data) {
        if ($iscount) {
            $result{$_} += $data->$_() for (keys %result);
            $result{total} += $data->$_() for (keys %result);
        } else {
            for my $k (keys %result) {
                $result{$k} = [ grep { $_ > 0 } split( ',', $data->$k()) ];
            }
        }
    }

    return \%result;
}



__PACKAGE__->register_method(
    method        => "checked_in_with_fines",
    api_name      => "sharestuff.actor.user.checked_in_with_fines",
    authoritative => 1,
    argc          => 2,
    signature     => q/@see sharestuff.actor.user.checked_out/
);

sub checked_in_with_fines {
	my( $self, $conn, $auth, $userid ) = @_;

	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;

	if( $userid ne $e->requestor->id ) {
		return $e->event unless $e->allowed('VIEW_CIRCULATIONS');
	}

	# money is owed on these items and they are checked in
	my $open = $e->search_action_circulation(
		{
			usr				=> $userid, 
			xact_finish		=> undef,
			checkin_time	=> { "!=" => undef },
		}
	);


	my( @lost, @cr, @lo );
	for my $c (@$open) {
		push( @lost, $c->id ) if $c->stop_fines eq 'LOST';
		push( @cr, $c->id ) if $c->stop_fines eq 'CLAIMSRETURNED';
		push( @lo, $c->id ) if $c->stop_fines eq 'LONGOVERDUE';
	}

	return {
		lost		=> \@lost,
		claims_returned	=> \@cr,
		long_overdue		=> \@lo
	};
}


sub _sigmaker {
    my ($api, $desc, $auth) = @_;
    $desc = $desc ? (" " . $desc) : '';
    my $ids = ($api =~ /ids$/) ? 1 : 0;
    my @sig = (
        argc      => 1,
        method    => "user_transaction_history",
        api_name  => "sharestuff.actor.user.transactions.$api",
        signature => {
            desc   => "For a given User ID, returns a list of billable transaction" .
                      ($ids ? " id" : '') .
                      "s$desc, optionally filtered by type and/or fields in money.billable_xact_summary.  " .
                      "The VIEW_USER_TRANSACTIONS permission is required to view another user's transactions",
            params => [
                {desc => 'Authentication token',        type => 'string'},
                {desc => 'User ID',                     type => 'number'},
                {desc => 'Transaction type (optional)', type => 'number'},
                {desc => 'Hash of Billable Transaction Summary filters (optional)', type => 'object'}
            ],
            return => {
                desc => 'List of transaction' . ($ids ? " id" : '') . 's, Event on error'
            },
        }
    );
    $auth and push @sig, (authoritative => 1);
    return @sig;
}

my %auth_hist_methods = (
    'history'             => '',
    'history.have_charge' => 'that have an initial charge',
    'history.still_open'  => 'that are not finished',
    'history.have_balance'         => 'that have a balance',
    'history.have_bill'            => 'that have billings',
    'history.have_bill_or_payment' => 'that have non-zero-sum billings or at least 1 payment',
    'history.have_payment' => 'that have at least 1 payment',
);

foreach (keys %auth_hist_methods) {
    __PACKAGE__->register_method(_sigmaker($_,       $auth_hist_methods{$_}, 1));
    __PACKAGE__->register_method(_sigmaker("$_.ids", $auth_hist_methods{$_}, 1));
    __PACKAGE__->register_method(_sigmaker("$_.fleshed", $auth_hist_methods{$_}, 1));
}

sub user_transaction_history {
	my( $self, $conn, $auth, $userid, $type, $filter, $options ) = @_;
    $filter ||= {};
    $options ||= {};

	my $e = new_editor(authtoken=>$auth);
	return $e->die_event unless $e->checkauth;

	if ($e->requestor->id ne $userid) {
        return $e->die_event unless $e->allowed('VIEW_USER_TRANSACTIONS');
	}

	my $api = $self->api_name;
	my @xact_finish  = (xact_finish => undef ) if ($api =~ /history\.still_open$/);     # What about history.still_open.ids?

	if(defined($type)) {
		$filter->{'xact_type'} = $type;
	}

	if($api =~ /have_bill_or_payment/o) {

        # transactions that have a non-zero sum across all billings or at least 1 payment
        $filter->{'-or'} = {
            'balance_owed' => { '<>' => 0 },
            'last_payment_ts' => { '<>' => undef }
        };

    } elsif($api =~ /have_payment/) {

        $filter->{last_payment_ts} ||= {'<>' => undef};

    } elsif( $api =~ /have_balance/o) {

        # transactions that have a non-zero overall balance
        $filter->{'balance_owed'} = { '<>' => 0 };

	} elsif( $api =~ /have_charge/o) {

        # transactions that have at least 1 billing, regardless of whether it was voided
        $filter->{'last_billing_ts'} = { '<>' => undef };

	} elsif( $api =~ /have_bill/o) {    # needs to be an elsif, or we double-match have_bill_or_payment!

        # transactions that have non-zero sum across all billings.  This will exclude
        # xacts where all billings have been voided
        $filter->{'total_owed'} = { '<>' => 0 };
	}

    my $options_clause = { order_by => { mbt => 'xact_start DESC' } };
    $options_clause->{'limit'} = $options->{'limit'} if $options->{'limit'}; 
    $options_clause->{'offset'} = $options->{'offset'} if $options->{'offset'}; 

    my $mbts = $e->search_money_billable_transaction_summary(
        [   { usr => $userid, @xact_finish, %$filter },
            $options_clause
        ]
    );

    return [map {$_->id} @$mbts] if $api =~ /\.ids/;
    return $mbts unless $api =~ /fleshed/;

	my @resp;
	for my $t (@$mbts) {
			
		if( $t->xact_type ne 'circulation' ) {
			push @resp, {transaction => $t};
			next;
		}

        my $circ_data = flesh_circ($e, $t->id);
		push @resp, {transaction => $t, %$circ_data};
	}

	return \@resp; 
}


__PACKAGE__->register_method(
    method   => "retrieve_perms",
    api_name => "sharestuff.actor.permissions.retrieve",
    notes    => "Returns a list of permissions"
);
sub retrieve_perms {
	my( $self, $client ) = @_;
	return $apputils->simple_scalar_request(
		"sharestuff.cstore",
		"sharestuff.cstore.direct.permission.perm_list.search.atomic",
		{ id => { '!=' => undef } }
	);
}

__PACKAGE__->register_method(
    method   => "retrieve_groups",
    api_name => "sharestuff.actor.groups.retrieve",
    notes    => "Returns a list of user groups"
);
sub retrieve_groups {
	my( $self, $client ) = @_;
	return new_editor()->retrieve_all_permission_grp_tree();
}

__PACKAGE__->register_method(
    method   => "retrieve_groups_tree",
    api_name => "sharestuff.actor.groups.tree.retrieve",
    notes    => "Returns a list of user groups"
);
	
sub retrieve_groups_tree {
	my( $self, $client ) = @_;
	return new_editor()->search_permission_grp_tree(
		[
			{ parent => undef},
			{	
				flesh				=> -1,
				flesh_fields	=> { pgt => ["children"] }, 
				order_by			=> { pgt => 'name'}
			}
		]
	)->[0];
}


__PACKAGE__->register_method(
    method   => "add_user_to_groups",
    api_name => "sharestuff.actor.user.set_groups",
    notes    => "Adds a user to one or more permission groups"
);
	
sub add_user_to_groups {
	my( $self, $client, $authtoken, $userid, $groups ) = @_;

	my( $requestor, $target, $evt ) = $apputils->checkses_requestor(
		$authtoken, $userid, 'CREATE_USER_GROUP_LINK' );
	return $evt if $evt;

	( $requestor, $target, $evt ) = $apputils->checkses_requestor(
		$authtoken, $userid, 'REMOVE_USER_GROUP_LINK' );
	return $evt if $evt;

	my $perms = $apputils->simplereq(
		'sharestuff.cstore',
		'sharestuff.cstore.direct.permission.usr_grp_map.search.atomic', { usr => $userid } );
    for my $p (@$perms) {
		$apputils->simplereq(
    		'sharestuff.cstore',
	    	'sharestuff.cstore.direct.permission.usr_grp_map.delete', $p );
    }
		
	for my $group (@$groups) {
		my $link = Fieldmapper::permission::usr_grp_map->new;
		$link->grp($group);
		$link->usr($userid);

		my $id = $apputils->simplereq(
			'sharestuff.cstore',
			'sharestuff.cstore.direct.permission.usr_grp_map.create', $link );
	}

	return 1;
}

__PACKAGE__->register_method(
    method   => "get_user_perm_groups",
    api_name => "sharestuff.actor.user.get_groups",
    notes    => "Retrieve a user's permission groups."
);


sub get_user_perm_groups {
	my( $self, $client, $authtoken, $userid ) = @_;

	my( $requestor, $target, $evt ) = $apputils->checkses_requestor(
		$authtoken, $userid, 'VIEW_PERM_GROUPS' );
	return $evt if $evt;

	return $apputils->simplereq(
		'sharestuff.cstore',
		'sharestuff.cstore.direct.permission.usr_grp_map.search.atomic', { usr => $userid } );
}	

__PACKAGE__->register_method(
    method        => 'fetch_patron_note',
    api_name      => 'sharestuff.actor.note.retrieve.all',
    authoritative => 1,
    signature     => q/
		Returns a list of notes for a given user
		Requestor must have VIEW_USER permission if pub==false and
		@param authtoken The login session key
		@param args Hash of params including
			patronid : the patron's id
			pub : true if retrieving only public notes
	/
);

sub fetch_patron_note {
	my( $self, $conn, $authtoken, $args ) = @_;
	my $patronid = $$args{patronid};

	my($reqr, $evt) = $U->checkses($authtoken);
	return $evt if $evt;

	my $patron;
	($patron, $evt) = $U->fetch_user($patronid);
	return $evt if $evt;

	if($$args{pub}) {
		if( $patronid ne $reqr->id ) {
			$evt = $U->check_perms($reqr->id, $patron->home_ou, 'VIEW_USER');
			return $evt if $evt;
		}
		return $U->cstorereq(
			'sharestuff.cstore.direct.actor.usr_note.search.atomic', 
			{ usr => $patronid, pub => 't' } );
	}

	$evt = $U->check_perms($reqr->id, $patron->home_ou, 'VIEW_USER');
	return $evt if $evt;

	return $U->cstorereq(
		'sharestuff.cstore.direct.actor.usr_note.search.atomic', { usr => $patronid } );
}

__PACKAGE__->register_method(
    method    => 'create_user_note',
    api_name  => 'sharestuff.actor.note.create',
    signature => q/
		Creates a new note for the given user
		@param authtoken The login session key
		@param note The note object
	/
);
sub create_user_note {
	my( $self, $conn, $authtoken, $note ) = @_;
	my $e = new_editor(xact=>1, authtoken=>$authtoken);
	return $e->die_event unless $e->checkauth;

	my $user = $e->retrieve_actor_user($note->usr)
		or return $e->die_event;

	return $e->die_event unless 
		$e->allowed('UPDATE_USER',$user->home_ou);

	$note->creator($e->requestor->id);
	$e->create_actor_usr_note($note) or return $e->die_event;
	$e->commit;
	return $note->id;
}


__PACKAGE__->register_method(
    method    => 'delete_user_note',
    api_name  => 'sharestuff.actor.note.delete',
    signature => q/
		Deletes a note for the given user
		@param authtoken The login session key
		@param noteid The note id
	/
);
sub delete_user_note {
	my( $self, $conn, $authtoken, $noteid ) = @_;

	my $e = new_editor(xact=>1, authtoken=>$authtoken);
	return $e->die_event unless $e->checkauth;
	my $note = $e->retrieve_actor_usr_note($noteid)
		or return $e->die_event;
	my $user = $e->retrieve_actor_user($note->usr)
		or return $e->die_event;
	return $e->die_event unless 
		$e->allowed('UPDATE_USER', $user->home_ou);
	
	$e->delete_actor_usr_note($note) or return $e->die_event;
	$e->commit;
	return 1;
}


__PACKAGE__->register_method(
    method    => 'update_user_note',
    api_name  => 'sharestuff.actor.note.update',
    signature => q/
		@param authtoken The login session key
		@param note The note
	/
);

sub update_user_note {
	my( $self, $conn, $auth, $note ) = @_;
	my $e = new_editor(authtoken=>$auth, xact=>1);
	return $e->die_event unless $e->checkauth;
	my $patron = $e->retrieve_actor_user($note->usr)
		or return $e->die_event;
	return $e->die_event unless 
		$e->allowed('UPDATE_USER', $patron->home_ou);
	$e->update_actor_user_note($note)
		or return $e->die_event;
	$e->commit;
	return 1;
}

__PACKAGE__->register_method(
    method    => 'usrname_exists',
    api_name  => 'sharestuff.actor.username.exists',
    signature => {
        desc  => 'Check if a username is already taken (by an undeleted patron)',
        param => [
            {desc => 'Authentication token', type => 'string'},
            {desc => 'Username',             type => 'string'}
        ],
        return => {
            desc => 'id of existing user if username exists, undef otherwise.  Event on error'
        },
    }
);

sub usrname_exists {
	my( $self, $conn, $auth, $usrname ) = @_;
	my $e = new_editor(authtoken=>$auth);
	return $e->event unless $e->checkauth;
	my $a = $e->search_actor_user({usrname => $usrname}, {idlist=>1});
	return $$a[0] if $a and @$a;
	return undef;
}

# Retain the old typo API name just in case
__PACKAGE__->register_method(
    method   => 'fetch_org_by_shortname',
    api_name => 'sharestuff.actor.org_unit.retrieve_by_shorname',
);
__PACKAGE__->register_method(
    method   => 'fetch_org_by_shortname',
    api_name => 'sharestuff.actor.org_unit.retrieve_by_shortname',
);
sub fetch_org_by_shortname {
	my( $self, $conn, $sname ) = @_;
	my $e = new_editor();
	my $org = $e->search_actor_org_unit({ shortname => uc($sname)})->[0];
	return $e->event unless $org;
	return $org;
}


__PACKAGE__->register_method(
    method   => 'session_home_lib',
    api_name => 'sharestuff.actor.session.home_lib',
);

sub session_home_lib {
	my( $self, $conn, $auth ) = @_;
	my $e = new_editor(authtoken=>$auth);
	return undef unless $e->checkauth;
	my $org = $e->retrieve_actor_org_unit($e->requestor->home_ou);
	return $org->shortname;
}

__PACKAGE__->register_method(
    method    => 'session_safe_token',
    api_name  => 'sharestuff.actor.session.safe_token',
    signature => q/
		Returns a hashed session ID that is safe for export to the world.
		This safe token will expire after 1 hour of non-use.
		@param auth Active authentication token
	/
);

sub session_safe_token {
	my( $self, $conn, $auth ) = @_;
	my $e = new_editor(authtoken=>$auth);
	return undef unless $e->checkauth;

	my $safe_token = md5_hex($auth);

	$cache ||= OpenSRF::Utils::Cache->new("global", 0);

	# Add more like the following if needed...
	$cache->put_cache(
		"safe-token-home_lib-shortname-$safe_token",
		$e->retrieve_actor_org_unit(
			$e->requestor->home_ou
		)->shortname,
		60 * 60
	);

	return $safe_token;
}


__PACKAGE__->register_method(
    method    => 'safe_token_home_lib',
    api_name  => 'sharestuff.actor.safe_token.home_lib.shortname',
    signature => q/
		Returns the home library shortname from the session
		asscociated with a safe token from generated by
		sharestuff.actor.session.safe_token.
		@param safe_token Active safe token
	/
);

sub safe_token_home_lib {
	my( $self, $conn, $safe_token ) = @_;

	$cache ||= OpenSRF::Utils::Cache->new("global", 0);
	return $cache->get_cache( 'safe-token-home_lib-shortname-'. $safe_token );
}


__PACKAGE__->register_method(
    method   => "user_retrieve_parts",
    api_name => "sharestuff.actor.user.retrieve.parts",
);

sub user_retrieve_parts {
	my( $self, $client, $auth, $user_id, $fields ) = @_;
	my $e = new_editor(authtoken => $auth);
	return $e->event unless $e->checkauth;
    $user_id ||= $e->requestor->id;
	if( $e->requestor->id != $user_id ) {
		return $e->event unless $e->allowed('VIEW_USER');
	}
	my @resp;
	my $user = $e->retrieve_actor_user($user_id) or return $e->event;
	push(@resp, $user->$_()) for(@$fields);
	return \@resp;
}


__PACKAGE__->register_method (
	method		=> 'verify_user_password',
	api_name	=> 'sharestuff.actor.verify_user_password',
	signature	=> q/
        Given a barcode or username and the MD5 encoded password, 
        returns 1 if the password is correct.  Returns 0 otherwise.
	/
);

sub verify_user_password {
    my($self, $conn, $auth, $username, $password) = @_;
    my $e = new_editor(authtoken => $auth);
	return $e->die_event unless $e->checkauth;

    my $user = $e->search_actor_user({usrname => $username})->[0] or return 0;
    return 0 if (!$user);
    return $e->event unless $e->allowed('VIEW_USER', $user->home_ou);
    return 1 if $user->passwd eq $password;
    return 0;
}

sub check_password_strength_default {
    my $password = shift;
    # Use the default set of checks
    if ( (length($password) < 7) or 
            ($password !~ m/.*\d+.*/) or 
            ($password !~ m/.*[A-Za-z]+.*/)
       ) {
        return 0;
    }
    return 1;
}

sub check_password_strength_custom {
    my ($password, $pw_regex) = @_;

    $pw_regex = qr/$pw_regex/;
    if ($password !~  /$pw_regex/) {
        return 0;
    }
    return 1;
}


1;
